"""Synthetic random price data feed for unit testing."""

import random
from decimal import Decimal
from typing import Dict, Optional, List, Iterable, Type

import pandas as pd
from tqdm_loggable.auto import tqdm

from eth_defi.event_reader.block_header import BlockHeader
from eth_defi.price_oracle.oracle import BasePriceOracle
from eth_defi.event_reader.reorganisation_monitor import ReorganisationMonitor

from .timeframe import Timeframe
from .trade_feed import TradeFeed, Trade
from .direct_feed_pair import PairId


class SyntheticTradeFeed(TradeFeed):
    """Synthetic trade feed that produces random trades."""

    def __init__(self,
                 pairs: List[PairId],
                 oracles: Dict[PairId, BasePriceOracle],
                 reorg_mon: ReorganisationMonitor,
                 data_retention_time: Optional[pd.Timedelta] = None,
                 random_seed = 1,
                 start_price_range=150,
                 end_price_range=300,
                 min_trades_per_block=0,
                 max_trades_per_block=10,
                 min_amount=50,
                 max_amount=50,
                 price_movement_per_trade=2.5,
                 timeframe: Timeframe = Timeframe("1min"),
                 prices: Dict[PairId, float] = None,
                 broken_wick_block_frequency=0,
                 ):
        super().__init__(
            pairs=pairs,
            oracles=oracles,
            reorg_mon=reorg_mon,
            data_retention_time=data_retention_time,
            timeframe=timeframe,
        )

        self.pairs = pairs
        self.min_trades_per_block = min_trades_per_block
        self.max_trades_per_block = max_trades_per_block
        self.min_amount = min_amount
        self.max_amount = max_amount
        self.price_movement_per_trade = price_movement_per_trade
        self.broken_wick_block_frequency = broken_wick_block_frequency
        self.random_gen = random.Random(random_seed)

        if not prices:
            self.prices = {
                p: self.random_gen.randint(start_price_range, end_price_range)
                for p in pairs
            }
        else:
            self.prices = prices

    def fetch_trades(self, start_block: int, end_block: Optional[int], tqdm: Optional[Type[tqdm]] = None) -> Iterable[Trade]:
        """Generate few random trades per block per pair."""

        block_data = {b.block_number: b for b in self.reorg_mon.fetch_block_data(start_block, end_block)}

        max_blocks = end_block - start_block

        if max_blocks > 5 and tqdm:
            progress_bar = tqdm(total=max_blocks)
        else:
            progress_bar = None

        for block_num in range(start_block, end_block + 1):
            for p in self.pairs:
                trades_per_block = self.random_gen.randint(self.min_trades_per_block, self.max_trades_per_block)
                for trade_idx in range(trades_per_block):

                    self.prices[p] += self.random_gen.uniform(-self.price_movement_per_trade, self.price_movement_per_trade)
                    self.prices[p] = max(self.prices[p], 0.00001)  # Don't go to negative prices
                    price = self.prices[p]

                    # Simulate broken data by inserting prices 1000x out of range
                    # See test_filter_wick.py
                    if self.broken_wick_block_frequency:
                        if self.random_gen.randint(0, self.broken_wick_block_frequency) == 0:
                            if self.random_gen.randint(0, 2) == 1:
                                price *= 1000
                            else:
                                price /= 1000

                    amount = self.random_gen.uniform(self.min_amount, self.max_amount)

                    block: BlockHeader = block_data[block_num]
                    tx_hash = hex(self.random_gen.randint(2**31, 2**32))
                    log_index = trade_idx

                    exchange_rate = self.get_exchange_rate(p)

                    yield Trade(
                           pair=p,
                           block_number=block_num,
                           block_hash=block.block_hash,
                           timestamp=pd.Timestamp.utcfromtimestamp(block.timestamp).tz_localize(None),
                           tx_hash=tx_hash,
                           log_index=log_index,
                           price=Decimal(price),
                           amount=Decimal(amount),
                           exchange_rate=exchange_rate,
                        )

            if progress_bar:
                progress_bar.update(1)

        if progress_bar:
            progress_bar.close()



"""Lending rates and lending protocols.

Data structures for understanding and using lending rates
like supply APR and borrow APR across various DeFi lending protocols.

See :py:class:`LendingReserveUniverse` on how to load the data.
"""
import warnings
from _decimal import Decimal
from enum import Enum
import datetime

from dataclasses import dataclass, field
from typing import TypeAlias, Tuple, Collection, Iterator, Dict, Set

from dataclasses_json import dataclass_json, config

import pandas as pd

from tradingstrategy.chain import ChainId
from tradingstrategy.token import Token
from tradingstrategy.types import UNIXTimestamp, PrimaryKey, TokenSymbol, Slug, NonChecksummedAddress, URL
from tradingstrategy.utils.groupeduniverse import PairGroupedUniverse

from eth_defi.aave_v3.rates import SECONDS_PER_YEAR


class LendingProtocolType(str, Enum):
    """Supported lending protocols."""
    aave_v3 = "aave_v3"


class LendingCandleType(str, Enum):
    """What kind of lending price feeds we have."""
    stable_borrow_apr = "stable_borrow_apr"
    variable_borrow_apr = "variable_borrow_apr"
    supply_apr = "supply_apr"


class UnknownLendingReserve(Exception):
    """Does not know about this lending reserve."""


class NoLendingData(Exception):
    """Lending data missing for asked period.

    Reserve was likely not active yet.
    """


@dataclass_json
@dataclass
class LendingReserveAdditionalDetails:
    """Additional details for a lending reserve."""

    #: Latest Loan-To-Value ratio
    ltv: float | None = None

    #: Latest liquidation threshold
    liquidation_threshold: float | None = None


@dataclass_json
@dataclass
class LendingReserve:
    """Describe data for one lending reserve.

    - This is mostly modelled by how lending on Aave works

    - When you lend and asset you will receive a correcting ponding amount of
      aToken in return

    - aToken ERC-20 contract has dynamic balanceOf() function that tells
      you how much principal + interest you have gained
    """

    #: Primary key to identity the lending reserve
    #: Use lending reserve universe to map this to chain id and a smart contract address
    reserve_id: PrimaryKey

    #: The slug of this lending reserve
    #:
    #: E.g. `aave-v3`.
    #:
    reserve_slug: Slug

    #: Which lending protocol this asset is for
    protocol_slug: LendingProtocolType

    #: The id on which chain this lending reserve is deployed
    chain_id: ChainId

    #: The slug on which chain this lending reserve is deployed.
    #:
    #: Needed for website URL linking.
    #:
    #: E.g. `polygon`.
    #:
    chain_slug: Slug

    #: The internal ID of this asset, this might be changed
    asset_id: PrimaryKey

    #: The asset name of this lending reserve
    asset_name: str

    #: The asset symbol of this lending reserve
    asset_symbol: TokenSymbol

    #: The ERC-20 address of the underlying asset
    asset_address: NonChecksummedAddress

    #: Number of decimals
    asset_decimals: int

    #: The internal ID of this the aToken, this might be changed
    atoken_id: PrimaryKey

    #: The aToken symbol of this lending reserve
    atoken_symbol: TokenSymbol

    #: The ERC-20 address of the aToken
    atoken_address: NonChecksummedAddress

    #: The decimals of the aToken .
    #:
    #: Should be always the same as :py:attr:`asset_decimals`
    #:
    atoken_decimals: int

    #: The internal ID of this the vToken, this might be changed
    vtoken_id: PrimaryKey = field(metadata=config(field_name="variable_debt_token_id"))

    #: The vToken symbol of this lending reserve
    vtoken_symbol: TokenSymbol = field(metadata=config(field_name="variable_debt_token_symbol"))

    #: The ERC-20 address of the vToken
    vtoken_address: NonChecksummedAddress = field(metadata=config(field_name="variable_debt_token_address"))

    #: The decimals of the aToken .
    vtoken_decimals: int = field(metadata=config(field_name="variable_debt_token_decimals"))

    #: Other details like latest LTV ratio or liquidation threshold
    additional_details: LendingReserveAdditionalDetails

    def __eq__(self, other: "LendingReserve") -> bool:
        assert isinstance(other, LendingReserve)
        return self.chain_id == other.chain_id and self.protocol_slug == other.protocol_slug and self.asset_address == other.asset_address

    def __hash__(self):
        return hash((self.chain_id, self.protocol_slug, self.asset_address))

    def __repr__(self):
        return f"<LendingReserve #{self.reserve_id} for asset {self.asset_symbol} ({self.asset_address}) in protocol {self.protocol_slug.name} on {self.chain_id.get_name()} >"
    
    def get_asset(self) -> Token:
        """Return description for the underlying asset."""
        return Token(
            self.chain_id,
            self.asset_symbol,
            self.asset_address,
            self.asset_decimals,
        )

    def get_atoken(self) -> Token:
        """Return description for aToken."""
        return Token(
            self.chain_id,
            self.atoken_symbol,
            self.atoken_address,
            self.atoken_decimals,
        )
    
    def get_vtoken(self) -> Token:
        """Return description for vToken (variable debt token)."""
        return Token(
            self.chain_id,
            self.vtoken_symbol,
            self.vtoken_address,
            self.vtoken_decimals,
        )

    def get_link(self) -> URL:
        """Get the market data page link"""
        return f"https://tradingstrategy.ai/trading-view/{self.chain_id.get_slug()}/lending/{self.protocol_slug}/{self.asset_symbol.lower()}"

#: How to symbolically identify a lending reserve.
#:
#: Used in human written code instead of unreadable smart contract addresses.
#:
#: - Chain id
#: - Lending protocol type
#: - Reserve token symbol
#: - (Optional) smart contract address
#:
#: Example: `(ChainId.polygon, LendingProtocolType.aave_v3, "USDC")`
#:
#: If there are multiple reserves with the same token, the fourth
#: parameter is a smart contract address that distinguishes these.
#: It is currently not used.
#:
#: Note that the underlying LendingReserve internal ids may change and slugs,
#: only smart contract addresses stay stable.
#:
LendingReserveDescription: TypeAlias = Tuple[ChainId, LendingProtocolType, TokenSymbol] | \
                                       Tuple[ChainId, LendingProtocolType, TokenSymbol, NonChecksummedAddress]



@dataclass_json
@dataclass
class LendingReserveUniverse:
    """Lending reserve universe contains metadata of all lending pools you can access.

    You can use reserve universe as a map to resolve symbolic information
    to the data primary keys.

    For the usage see :py:meth:`resolve_lending_reserve`.

    Example how to create:

    .. code-block:: python

        exchange_universe = client.fetch_exchange_universe()

        quote_tokens = {
            "0x2791bca1f2de4661ed88a30c99a7a9449aa84174",  # USDC polygon
            "0xc2132d05d31c914a87c6611c10748aeb04b58e8f",  # USDT polygon
        }

        pairs_df = client.fetch_pair_universe().to_pandas()

        # Find out all volatile pairs traded against USDC and USDT on Polygon
        pairs_df = filter_for_chain(pairs_df, ChainId.polygon)
        pairs_df = filter_for_stablecoins(pairs_df, StablecoinFilteringMode.only_volatile_pairs)
        pairs_df = filter_for_quote_tokens(pairs_df, quote_tokens)

        # Create lending universe and trading universe with the cross section of
        # - Available assets in the lending protocols
        # - Asset we can trade
        lending_reserves = client.fetch_lending_reserve_universe()
        lending_only_pairs_df = filter_for_base_tokens(pairs_df, lending_reserves.get_asset_addresses())

        pair_universe = PandasPairUniverse(lending_only_pairs_df, exchange_universe=exchange_universe)

        # Lending reserves have around ~320 individual trading pairs on Polygon across different DEXes
        assert 1 < pair_universe.get_count() < 1_000

        eth_usdc = pair_universe.get_pair_by_human_description((ChainId.polygon, "uniswap-v3", "WETH", "USDC"))
        matic_usdc = pair_universe.get_pair_by_human_description((ChainId.polygon, "uniswap-v3", "WMATIC", "USDC"))

        # Random token not in the Aave v3 supported reserves
        full_pair_universe =  PandasPairUniverse(pairs_df, exchange_universe=exchange_universe)
        bob_usdc = full_pair_universe.get_pair_by_human_description((ChainId.polygon, "uniswap-v3", "BOB", "USDC"))

        assert lending_reserves.can_leverage(eth_usdc.get_base_token())
        assert lending_reserves.can_leverage(matic_usdc.get_base_token())
        assert not lending_reserves.can_leverage(bob_usdc.get_base_token())

    """

    #: Reserve ID -> Reserve data mapping
    reserves: dict[PrimaryKey, LendingReserve]

    def __repr__(self):
        return f"<LendingReserveUniverse with {len(self.reserves)} reserves>"

    def get_count(self) -> int:
        """How many reserves we have.

        :return:
            Number of lending reserves in the trading universe
        """
        return len(self.reserves)

    def iterate_reserves(self) -> Iterator[LendingReserve]:
        return self.reserves.values()

    def get_reserve_by_id(self, reserve_id: PrimaryKey) -> LendingReserve | None:
        return self.reserves.get(reserve_id)

    def get_reserve_by_symbol_and_chain(
            self,
            token_symbol: TokenSymbol,
            chain_id: ChainId,
    ) -> LendingReserve | None:
        warnings.warn("get_reserve_by_symbol_and_chain() has been deprecated in the favour of get_by_chain_and_symbol()", DeprecationWarning, stacklevel=2)
        return self.get_by_chain_and_symbol(chain_id, token_symbol)

    def get_by_chain_and_symbol(
            self,
            chain_id: ChainId,
            token_symbol: TokenSymbol,
    ) -> LendingReserve:
        """Fetch a specific lending reserve.

        :raise UnknownLendingReserve:
            If we do not have data available.
        """

        assert isinstance(chain_id, ChainId), f"Expected chain_id, got {chain_id.__class__}: {chain_id}"

        for reserve in self.reserves.values():
            if reserve.asset_symbol == token_symbol and reserve.chain_id == chain_id:
                return reserve

        raise UnknownLendingReserve(f"Could not find lending reserve {chain_id}: {token_symbol}. We have {len(self.reserves)} reserves loaded.")

    def get_by_chain_and_address(
            self,
            chain_id: ChainId,
            asset_address: NonChecksummedAddress,
    ) -> LendingReserve:
        """Get a lending reserve by persistent data.

        :raise UnknownLendingReserve:
            If we do not have data available.
        """

        assert isinstance(chain_id, ChainId), f"Got: {chain_id}"

        for reserve in self.reserves.values():
            if reserve.asset_address == asset_address and reserve.chain_id == chain_id:
                return reserve

        raise UnknownLendingReserve(f"Could not find lending reserve on chain {chain_id.get_name()}, reserve token address {asset_address}. We have {len(self.reserves)} reserves loaded.")

    def limit(self, reserve_descriptions: Collection[LendingReserveDescription]) -> "LendingReserveUniverse":
        """Remove all lending reverses that are not on the whitelist.

        Used to reduce the lending reserve universe to wanted tradeable assets
        after loading is done.
        """
        new_reserves = {}

        for d in reserve_descriptions:
            r = self.resolve_lending_reserve(d)
            new_reserves[r.reserve_id] = r

        return LendingReserveUniverse(new_reserves)

    def limit_to_chain(self, chain_id: ChainId) -> "LendingReserveUniverse":
        """Drop all lending reserves except ones on a specific chain."""
        assert isinstance(chain_id, ChainId)
        new_reserves = {r.reserve_id: r for r in self.reserves.values() if r.chain_id == chain_id}
        return LendingReserveUniverse(new_reserves)

    def limit_to_assets(self, assets: Set[TokenSymbol]) -> "LendingReserveUniverse":
        """Drop all lending reserves except listed tokens."""
        for a in assets:
            assert type(a) == str
        new_reserves = {r.reserve_id: r for r in self.reserves.values() if r.asset_symbol in assets}

        assert len(assets) == len(new_reserves), f"Could not resolve all assets: {assets}"

        return LendingReserveUniverse(new_reserves)

    def resolve_lending_reserve(self, reserve_decription: LendingReserveDescription) -> LendingReserve:
        """Looks up a lending reserve by a data match.

        Example:

        .. code-block:: python

            usdt_reserve = universe.resolve_lending_reserve(
                (ChainId.polygon,
                LendingProtocolType.aave_v3,
                "USDT")
            )
            assert isinstance(usdt_reserve, LendingReserve)
            assert usdt_reserve.asset_address == '0xc2132d05d31c914a87c6611c10748aeb04b58e8f'
            assert usdt_reserve.asset_symbol == "USDT"
            assert usdt_reserve.asset_name == '(PoS) Tether USD'
            assert usdt_reserve.asset_decimals == 6
            assert usdt_reserve.atoken_symbol == "aPolUSDT"
            assert usdt_reserve.atoken_decimals == 6

        :param reserve_decription:
            Human-readable tuple to resolve the lending reserve.

        :return:
            Metadata for this lending reserve

        :raise UnknownLendingReserve:
            If the loaded data does not contain the reserve
        """

        assert type(reserve_decription) == tuple, f"Lending reserve must be described as tuple, got {reserve_decription}"
        chain_id, slug, symbol, *optional = reserve_decription

        # Validate hard-coded inputs
        assert isinstance(chain_id, ChainId), f"Got {chain_id}"
        assert isinstance(slug, LendingProtocolType), f"Got {slug}"

        assert not optional, "Unsupported"

        for reserve in self.reserves.values():
            if reserve.chain_id == chain_id and \
                    reserve.protocol_slug == slug and \
                    reserve.asset_symbol == symbol:
                return reserve

        raise UnknownLendingReserve(f"Could not find lending reserve {reserve_decription}. We have {len(self.reserves)} reserves loaded.")

    def get_asset_addresses(self) -> Set[NonChecksummedAddress]:
        """Get all the token addresses in this dataset.

        A shortcut method.

        :return:
            Set of all assets in all lending reserves.
        """
        return {a.asset_address for a in self.reserves.values()}

    def can_leverage(self, token: Token) -> bool:
        """Can we go short/long on a specific token."""
        try:
            self.get_by_chain_and_address(
                token.chain_id,
                token.address,
            )
            return True
        except UnknownLendingReserve:
            return False

@dataclass_json
@dataclass
class LendingCandle:
    """Data structure presenting one OHLC lending candle"""

    #: Primary key to identity the trading pair
    #: Use lending reserve universe to map this to chain id and token symbol.
    #:
    #: See :py:class:`LendingReserveUniverse` how to look up this value.
    #:
    reserve_id: PrimaryKey

    #: Open timestamp for this candle.
    #: Note that the close timestamp you need to supply yourself based on the context.
    timestamp: UNIXTimestamp

    #: OHLC core data
    open: float

    #: OHLC core data
    close: float

    #: OHLC core data
    high: float

    #: OHLC core data
    low: float

    #: Schema definition for :py:class:`pd.DataFrame:
    #:
    #: Defines Pandas datatypes for columns in our candle data format.
    #: Useful e.g. when we are manipulating JSON/hand-written data.
    #:
    DATAFRAME_FIELDS = dict([
        ("reserve_id", "int"),
        ("timestamp", "datetime64[s]"),
        ("open", "float"),
        ("close", "float"),
        ("high", "float"),
        ("low", "float"),
    ])

    def __repr__(self):
        human_timestamp = datetime.utcfromtimestamp(self.timestamp)
        return f"@{human_timestamp} O:{self.open} H:{self.high} L:{self.low} C:{self.close}"

    @classmethod
    def convert_web_candles_to_dataframe(cls, web_candles: list[dict]) -> pd.DataFrame:
        """Return Pandas dataframe presenting candle data."""

        df = pd.DataFrame(web_candles)
        df = df.rename(columns={
            "ts": "timestamp",
            "o": "open",
            "h": "high",
            "l": "low",
            "c": "close",
        })
        df = df.astype(cls.DATAFRAME_FIELDS)

        # Convert unix timestamps to Pandas
        df["timestamp"] = pd.to_datetime(df["timestamp"])

        # Assume candles are always indexed by their timestamp
        df.set_index("timestamp", inplace=True, drop=True)

        return df


class LendingMetricUniverse(PairGroupedUniverse):
    """Single metric for multiple lending reserves.

    E.g. supply APR for both USDC and USDT.
    Internally handled as GroupBy :py:class:`pd.DataFrame`.

    This is designed so that you can easily develop strategies
    that access lending rates of multiple reserves.

    See also :py:class:`LendingCandleUniverse`.

    To access the data see

    - :py:meth:`tradingstrategy.utils.groupeduniverse.PairGroupedUniverse.get_single_pair_data`

    - :py:meth:`tradingstrategy.utils.groupeduniverse.PairGroupedUniverse.get_samples_by_pair`

    .. note ::

        No forward will enabled yet similar to :py:class:`tradingstrategy.candle.CandleUniverse`.

    """
    def __init__(self, df: pd.DataFrame, reserves: LendingReserveUniverse):
        """
        :param df:
            Output from client lending reserve dat aload

        :param reserves:
            Map of reserve metadata that helps us to resolve symbolic information.

        """
        self.reserves = reserves
        # Create a GroupBy universe from raw loaded lending reserve data
        # We do not need fix any wicks, because lending rates are not subject
        # to similar manipulation as DEX prices
        return super().__init__(
            df,
            primary_key_column="reserve_id",
            fix_wick_threshold=None,
        )

    def get_rates_by_reserve(
            self,
            reserve_description: LendingReserveDescription | LendingReserve
    ) -> pd.DataFrame:
        """Get all historical rates for a single pair.

        Example:

        .. code-block:: python

            rates = dataset.lending_candles.supply_apr.get_rates_by_reserve(
                (ChainId.polygon, LendingProtocolType.aave_v3, "USDC")
            )
            assert rates["open"][pd.Timestamp("2023-01-01")] == pytest.approx(1.836242)
            assert rates["close"][pd.Timestamp("2023-01-01")] == pytest.approx(1.780513)
        """
        if isinstance(reserve_description, LendingReserve):
            reserve = reserve_description
        elif type(reserve_description) == tuple:
            reserve = self.reserves.resolve_lending_reserve(reserve_description)
        else:
            raise AssertionError(f"Unknown lending reserve description: {reserve_description}")
        return self.get_samples_by_pair(reserve.reserve_id)

    def get_rates_by_id(self, reserve_id: PrimaryKey) -> pd.DataFrame:
        """Return lending rates for for a particular reserve."""
        return self.get_samples_by_pair(reserve_id)

    def get_single_rate(
        self,
        reserve: LendingReserve,
        when: pd.Timestamp | datetime.datetime,
        data_lag_tolerance: pd.Timedelta,
        kind="close",
    ) -> Tuple[float, pd.Timedelta]:
        """Get a single historical value of a lending rate.

        See :py:meth:`tradingstrategy.utils.groupeduniverse.PairGroupedUniverse.get_single_value`
        for documentation.

        Example:

        .. code-block:: python

            lending_reserves = client.fetch_lending_reserve_universe()

            usdt_desc = (ChainId.polygon, LendingProtocolType.aave_v3, "USDT")
            usdc_desc = (ChainId.polygon, LendingProtocolType.aave_v3, "USDC")

            lending_reserves = lending_reserves.limit([usdt_desc, usdc_desc])

            lending_candle_type_map = client.fetch_lending_candles_for_universe(
                lending_reserves,
                TimeBucket.d1,
                start_time=pd.Timestamp("2023-01-01"),
                end_time=pd.Timestamp("2023-02-01"),
            )

            lending_candles = LendingCandleUniverse(lending_candle_type_map, lending_reserves)

            usdc_reserve = lending_reserves.resolve_lending_reserve(usdc_desc)

            rate, lag = lending_candles.variable_borrow_apr.get_single_rate(
                usdc_reserve,
                pd.Timestamp("2023-01-01"),
                data_lag_tolerance=pd.Timedelta(days=1),
                kind="open",
            )
            assert rate == pytest.approx(1.836242)
            assert lag == ZERO_TIMEDELTA

        :return:
            Tuple (lending rate, data lag)

            The lending rate may be 0 if the reserve is special
            and borrowing/lending is disabled.
        """

        assert isinstance(reserve, LendingReserve), f"Got {reserve.__class__}"

        asset_name = reserve.asset_name
        link = reserve.get_link()

        return self.get_single_value(
            reserve.reserve_id,
            when,
            data_lag_tolerance,
            kind,
            asset_name=asset_name,
            link=link,
        )

    def estimate_accrued_interest(
        self,
        reserve: LendingReserveDescription | LendingReserve,
        start: datetime.datetime | pd.Timestamp,
        end: datetime.datetime | pd.Timestamp,
    ) -> Decimal:
        """Estimate how much credit or debt interest we would gain on Aave at a given period.

        Example:

        .. code-block:

            lending_reserves = client.fetch_lending_reserve_universe()

            usdc_desc = (ChainId.polygon, LendingProtocolType.aave_v3, "USDC")

            lending_reserves = lending_reserves.limit([usdc_desc])

            lending_candle_type_map = client.fetch_lending_candles_for_universe(
                lending_reserves,
                TimeBucket.d1,
                start_time=pd.Timestamp("2022-09-01"),
                end_time=pd.Timestamp("2023-09-01"),
            )
            lending_candles = LendingCandleUniverse(lending_candle_type_map, lending_reserves)

            # Estimate borrow cost
            borrow_interest_multiplier = lending_candles.variable_borrow_apr.estimate_accrued_interest(
                usdc_desc,
                start=pd.Timestamp("2022-09-01"),
                end=pd.Timestamp("2023-09-01"),
            )
            assert borrow_interest_multiplier == pytest.approx(Decimal('1.028597760665127969909038441'))

            # Estimate borrow cost
            supply_interest_multiplier = lending_candles.supply_apr.estimate_accrued_interest(
                usdc_desc,
                start=pd.Timestamp("2022-09-01"),
                end=pd.Timestamp("2023-09-01"),
            )
            assert supply_interest_multiplier == pytest.approx(Decimal('1.017786465640168974688961612'))

        :param reserve:
            Asset we are interested in.

        :param start:
            Start of the period

        :param end:
            End of the period

        :return:
            Interest multiplier.

            Multiply the starting balance with this number to get the interest applied balance at ``end``.

            1.0 = no interest.
        """

        if isinstance(start, datetime.datetime):
            start = pd.Timestamp(start)

        if isinstance(end, datetime.datetime):
            end = pd.Timestamp(end)

        assert isinstance(start, pd.Timestamp), f"Not a timestamp: {start}"
        assert isinstance(end, pd.Timestamp), f"Not a timestamp: {end}"

        assert start <= end

        df = self.get_rates_by_reserve(reserve)

        # TODO: Can we use index here to speed up?
        candles = df[(df["timestamp"] >= start) & (df["timestamp"] <= end)]

        if len(candles) == 0:
            total_candles = len(df)
            earliest = df.index[0]
            last = df.index[-1]

            if start < earliest or end > last:
                raise NoLendingData(
                    f"No lending data for asked period\n"
                    f"Reserve: {reserve}\n"
                    f"Asked interest rates for period {start} - {end}\n"
                    f"This reserve has {total_candles:,} total candles\n"
                    f"First candle is at is {earliest}, last candle is at {last}\n"
                )

            # We are within the available data period,
            # but we do not have candles.
            # This is because there are zero events generated in Aave during the period
            # and we are using sparse data.
            # We just use the last APR.
            ffill_indexer = df.index.get_indexer([start], method="ffill")
            before_match_iloc = ffill_indexer[0]
            avg_apr = Decimal(df["close"][before_match_iloc])

        else:
            # Calculate interest rate estimate based on the data

            # get average APR from high and low
            avg = candles[["high", "low"]].mean(axis=1)
            avg_apr = Decimal(avg.mean() / 100)

        duration = Decimal((end - start).total_seconds())
        accrued_interest_estimation = 1 + (1 * avg_apr) * duration / SECONDS_PER_YEAR  # Use Aave v3 seconds per year

        assert accrued_interest_estimation >= 1, f"Aave interest cannot be negative: {accrued_interest_estimation}"

        return accrued_interest_estimation


@dataclass
class LendingCandleUniverse:
    """Multiple metrics for all lending reserves.

    Track both lending and borrow rates, so it is
    easy to do credit supply, shorting and longing in the same trading strategy.

    See also :py:class:`LendingMetricUniverse`.

    Example:

    .. code-block:: python

        universe = client.fetch_lending_reserve_universe()

        usdt_desc = (ChainId.polygon, LendingProtocolType.aave_v3, "USDT")
        usdc_desc = (ChainId.polygon, LendingProtocolType.aave_v3, "USDC")

        limited_universe = universe.limit([usdt_desc, usdc_desc])

        usdt_reserve = limited_universe.resolve_lending_reserve(usdt_desc)
        usdc_reserve = limited_universe.resolve_lending_reserve(usdc_desc)

        lending_candle_type_map = client.fetch_lending_candles_for_universe(
            limited_universe,
            TimeBucket.d1,
            start_time=pd.Timestamp("2023-01-01"),
            end_time=pd.Timestamp("2023-02-01"),
        )

        universe = LendingCandleUniverse(lending_candle_type_map)

        # Read all data for a single reserve
        usdc_variable_borrow = universe.variable_borrow_apr.get_samples_by_pair(usdc_reserve.reserve_id)

        #            reserve_id      open      high       low     close  timestamp
        # timestamp
        # 2023-01-01           3  1.836242  1.839224  1.780513  1.780513 2023-01-01

        assert usdc_variable_borrow["open"][pd.Timestamp("2023-01-01")] == pytest.approx(1.836242)
        assert usdc_variable_borrow["close"][pd.Timestamp("2023-01-01")] == pytest.approx(1.780513)

        # Read data for multiple reserves for a time range

        #             reserve_id      open      high       low     close  timestamp
        # timestamp
        # 2023-01-05           6  2.814886  2.929328  2.813202  2.867843 2023-01-05
        # 2023-01-06           6  2.868013  2.928622  2.829608  2.866523 2023-01-06

        start = pd.Timestamp("2023-01-05")
        end = pd.Timestamp("2023-01-06")
        iterator = universe.supply_apr.iterate_samples_by_pair_range(start, end)
        for reserve_id, supply_apr in iterator:
            # Read supply apr only for USDT
            if reserve_id == usdt_reserve.reserve_id:
                assert len(supply_apr) == 2  # 2 days
                assert supply_apr["open"][pd.Timestamp("2023-01-05")] == pytest.approx(2.814886)
                assert supply_apr["close"][pd.Timestamp("2023-01-06")] == pytest.approx(2.866523)
    """

    stable_borrow_apr: LendingMetricUniverse | None = None
    variable_borrow_apr: LendingMetricUniverse | None = None
    supply_apr: LendingMetricUniverse | None = None

    def __init__(self, candle_type_dfs: Dict[LendingCandleType, pd.DataFrame], lending_reserve_universe: LendingReserveUniverse):
        """Create the lending candles universe.

        :param candle_type_dfs:
            Different lending metrics.

            Result from :py:meth:`tradingstrategy.client.Client.fetch_lending_candles_for_universe`.
        """

        if LendingCandleType.stable_borrow_apr in candle_type_dfs:
            self.stable_borrow_apr = LendingMetricUniverse(candle_type_dfs[LendingCandleType.stable_borrow_apr], lending_reserve_universe)

        if LendingCandleType.variable_borrow_apr in candle_type_dfs:
            self.variable_borrow_apr = LendingMetricUniverse(candle_type_dfs[LendingCandleType.variable_borrow_apr], lending_reserve_universe)

        if LendingCandleType.supply_apr in candle_type_dfs:
            self.supply_apr = LendingMetricUniverse(candle_type_dfs[LendingCandleType.supply_apr], lending_reserve_universe)

    @property
    def lending_reserves(self) -> LendingReserveUniverse:
        """Get the lending reserve universe.

        It is paired with each metric.
        """
        for metric in (self.stable_borrow_apr, self.variable_borrow_apr, self.supply_apr):
            if metric:
                return metric.reserves

        raise AssertionError("Empty LendingCandlesUniverse")


def convert_interest_rates_to_lending_candle_type_map(*args):
    """Convert lending and supply interest rates for all assets to a single lending_candle_type map.
    
    :param *args:
        List of dictionaries for each reserve. Each dictionary must have 3 keys/value pairs of the form:
        1. reserve_id: reserve id
        2. lending_data: pandas Series of lending rates
        3. supply_data: pandas Series of suppy rates

        Note that lending_data and supply_data should start and end at the same dates.

    :return: Dictionary of lending_candle_type_map
    """

    data = []
    for dictionary in args:
        assert set(dictionary.keys()) == {"reserve_id", "lending_data", "supply_data"}

        reserve_id = dictionary["reserve_id"]
        _lending_data = dictionary["lending_data"]
        _supply_data = dictionary["supply_data"]
        
        assert len(_lending_data) == len(_supply_data), "Lending data and supply data must have the same length"
        assert isinstance(_lending_data.index, pd.DatetimeIndex), "Index must be a DatetimeIndex"
        assert _lending_data.index.equals(_supply_data.index), "Lending data and supply data must have the same index"

        data.extend(zip(dictionary["lending_data"], dictionary["supply_data"], [reserve_id] * len(_lending_data), _lending_data.index))

    lending_candle_type_map = {
        LendingCandleType.variable_borrow_apr: pd.DataFrame({
            "open": [_data[0] for _data in data],
            "close": [_data[0] for _data in data],
            "high": [_data[0] for _data in data],
            "low": [_data[0] for _data in data],
            "timestamp": [_data[3] for _data in data],
            "reserve_id": [_data[2] for _data in data],
        }),
        LendingCandleType.supply_apr: pd.DataFrame({
            "open": [_data[1] for _data in data],
            "close": [_data[1] for _data in data],
            "high": [_data[1] for _data in data],
            "low": [_data[1] for _data in data],
            "timestamp": [_data[3] for _data in data],
            "reserve_id": [_data[2] for _data in data],
        }),
    }

    return lending_candle_type_map



"""Helper class to manage trading universes."""
import warnings
from dataclasses import dataclass
from typing import Tuple, Set, Optional

import pandas as pd

from tradingstrategy.candle import GroupedCandleUniverse
from tradingstrategy.chain import ChainId
from tradingstrategy.exchange import Exchange, ExchangeUniverse, ExchangeNotFoundError
from tradingstrategy.lending import LendingReserveUniverse, LendingCandleUniverse
from tradingstrategy.liquidity import GroupedLiquidityUniverse, ResampledLiquidityUniverse
from tradingstrategy.pair import PandasPairUniverse
from tradingstrategy.timebucket import TimeBucket


@dataclass
class Universe:
    """Trading universe description.

    Encapsulates all the data we need to make trading decisions.
    A trading strategy or a research notebook can use this
    class to pass around information of its available data.

    This includes

    - blockchains

    - exchanges

    - Trading pairs

    - Lending reserves

    - OHLCV data

    - Liquidity data

    - Lending data

    - Data timeframes (see :py:attr:`time_bucket`)
    """

    #: OHLCV data granularity
    time_bucket: TimeBucket

    #: List of blockchains the strategy trades on
    chains: Set[ChainId]

    #: All the exchanges for this strategy
    #:
    #: Presented with a capsulated :py:class:`ExchangeUniverse` that
    #: offers some convience methods.
    #:
    exchange_universe: Optional[ExchangeUniverse] = None

    #: List of trading pairs the strategy trades on
    pairs: Optional[PandasPairUniverse] = None

    #: Historical candles for the decision making
    candles: Optional[GroupedCandleUniverse] = None

    #: Liquidity data granularity
    liquidity_time_bucket: Optional[TimeBucket] = None

    #: Historical liquidity samples.
    #:
    #: Might not be loaded if the strategy does not need to access
    #: liquidity data.
    liquidity: Optional[GroupedLiquidityUniverse] = None

    #: Historical liquidity samples, resampled for backtesting speed.
    #:
    #: As strategies often do not need accurate liquidity information,
    #: approximation is enough, this resampled liquidity is
    #: optimised for backtesting speed.
    #:
    resampled_liquidity: Optional[ResampledLiquidityUniverse] = None

    #: Lending rates
    #:
    lending_candles: Optional[LendingCandleUniverse] = None

    #: List of exchanges the strategy trades on.
    #:
    #: TODO: Do not use this - will be be deprecated in the favour of :py:attr:`exchange_universe`
    exchanges: Optional[Set[Exchange]] = None

    def __post_init__(self):
        """Check that the constructor was called correctly.

        - Bail out the library user early for any error

        - Clean up some old code mess
        """
        if self.candles is not None:
            assert isinstance(self.candles, GroupedCandleUniverse), f"Expected GroupedCandleUniverse, got {self.candles.__class__}"

        if self.pairs is not None:
            assert isinstance(self.pairs, PandasPairUniverse), f"Expected PandasPairUniverse, got {self.pairs.__class__}"

        if self.exchanges is not None:
            # TODO: Legacy
            # Deprecate in some point
            if type(self.exchanges) == list:
                self.exchanges = set(self.exchanges)
            assert isinstance(self.exchanges, set), f"Expected set, got {self.exchanges.__class__}"

        if self.exchange_universe is not None:
            assert isinstance(self.exchange_universe, ExchangeUniverse), f"Expected dict, got {self.exchanges.__class__}"
        else:
            # TODO: Legacy backwards compatible
            self.exchange_universe = ExchangeUniverse({e.exchange_id: e for e in self.exchanges})

        if self.lending_candles is not None:
            assert isinstance(self.lending_candles, LendingCandleUniverse), f"Expected LendingCandleUniverse, got {self.exchanges.__class__}"

        assert type(self.chains) in (set, tuple, list), f"Expected collection of chains, got {self.chains.__class__}: {self.chains}"

        for c in self.chains:
            assert isinstance(c, ChainId), f"Expected ChainId object, got {c.__class__}: {c}"

    @property
    def lending_reserves(self) -> LendingReserveUniverse | None:
        """Each lending metric is paired with a copy of the universe

        :return:
            ``None`` if lending data is not loaded
        """
        if self.lending_candles is not None:
            return self.lending_candles.lending_reserves
        return None

    def get_candle_availability(self) -> Tuple[pd.Timestamp, pd.Timestamp]:
        """Get the time range for which we have candle data.

        Useful to check if the data is out of date.

        :return: start,end range
        """
        warnings.warn(f"use universe.candles.get_timestamp_range() instead", DeprecationWarning, stacklevel=2)
        return self.candles.get_timestamp_range()

    def get_single_exchange(self) -> Exchange:
        """For strategies that use only one exchange, get the exchange instance.

        :raise: AssertationError if multiple exchanges preset
        """
        warnings.warn(f"use universe.exchange_universe.get_single() instead", DeprecationWarning, stacklevel=2)
        assert self.exchange_universe.get_exchange_count() == 1
        return self.exchange_universe.get_single()

    def get_exchange_by_id(self, id: int) -> Optional[Exchange]:
        """Get exchange by its id.

        TODO: Replace exchange set with a proper exchange universe.

        :raise: ExchangeNotFoundError if exchange not found

        :return: Exchange instance
        """
        warnings.warn(f"use universe.exchange_universe.get_by_id() instead", DeprecationWarning, stacklevel=2)
        for exc in self.exchanges:
            if exc.exchange_id == id:
                return exc
        
        raise ExchangeNotFoundError(exchange_id=id)






"""Get candlestick price and volume data from Binance.
"""

import requests
import datetime
import pandas as pd
import numpy as np
import logging
import shutil

from tradingstrategy.timebucket import TimeBucket
from pathlib import Path
from tradingstrategy.utils.time import generate_monthly_timestamps, naive_utcnow, naive_utcfromtimestamp
from tradingstrategy.utils.groupeduniverse import resample_series


logger = logging.getLogger(__name__)


class BinanceDataFetchError(ValueError):
    """Something wrong with Binance."""


class BinanceDownloader:
    """Class for downloading Binance candlestick OHLCV data."""

    def __init__(self, cache_directory: Path = Path("/tmp/binance_data")):
        """Initialize BinanceCandleDownloader and create folder for cached data if it does not exist."""
        cache_directory.mkdir(parents=True, exist_ok=True)
        self.cache_directory = cache_directory

    def fetch_candlestick_data(
        self,
        symbol: str,
        time_bucket: TimeBucket,
        start_at: datetime.datetime,
        end_at: datetime.datetime,
        force_redownload=False,
    ):
        """Get clean candlestick price and volume data from Binance. If saved, use saved version, else create saved version.

        Note, if you want to use this data in our framework, you will need to add informational columns to the dataframe and overwrite it. See code below.

        .. code-block:: python
            symbol = "ETHUSDT"
            df = get_binance_candlestick_data(symbol, TimeBucket.h1, datetime.datetime(2021, 1, 1), datetime.datetime(2021, 4, 1))
            df = add_informational_columns(df, pair, EXCHANGE_SLUG)
            path = get_parquet_path(symbol, TimeBucket.h1, datetime.datetime(2021, 1, 1), datetime.datetime(2021, 4, 1))
            df.to_parquet(path)

        :param symbol:
            Trading pair symbol E.g. ETHUSDC

        :param interval:
            Can be one of `1s, 1m, 3m, 5m, 15m, 30m, 1h, 2h, 4h, 6h, 8h, 12h, 1d, 3d, 1w, 1M`

        :param start_at:
            Start date of the data

        :param end_at:
            End date of the data

        :param force_redownload:
            Force redownload of data from Binance and overwrite cached version

        :return:
            Pandas dataframe with the OHLCV data for the columns and datetimes as the index
        """
        if not force_redownload:
            try:
                return self.get_data_parquet(symbol, time_bucket, start_at, end_at)
            except:
                pass

        # to include the end date, we need to add one day
        end_at = end_at + datetime.timedelta(days=1)
        df = self._fetch_candlestick_data(symbol, time_bucket, start_at, end_at)
        end_at = end_at - datetime.timedelta(days=1)

        # write to parquet
        path = self.get_parquet_path(symbol, time_bucket, start_at, end_at)
        df.to_parquet(path)

        return df

    def _fetch_candlestick_data(
        self,
        symbol: str,
        time_bucket: TimeBucket,
        start_at: datetime.datetime,
        end_at: datetime.datetime,
    ):

        interval = get_binance_interval(time_bucket)

        params_str = f"symbol={symbol}&interval={interval}"

        if start_at:
            assert end_at, "If you specify a start_at, you must also specify an end_at"
            assert isinstance(
                start_at, datetime.datetime
            ), "start_at must be a datetime.datetime object"
            assert isinstance(
                end_at, datetime.datetime
            ), "end_at must be a datetime.datetime object"
            start_timestamp = int(start_at.timestamp() * 1000)
            end_timestamp = int(end_at.timestamp() * 1000)

        # generate timestamps for each iteration
        dates = [start_at]
        current_date = start_at
        while current_date < end_at:
            if (end_at - current_date) / time_bucket.to_timedelta() > 999:
                dates.append((current_date + time_bucket.to_timedelta() * 999))
                current_date += time_bucket.to_timedelta() * 999
            else:
                dates.append(end_at)
                current_date = end_at

        timestamps = [int(date.timestamp() * 1000) for date in dates]
        open_prices, high_prices, low_prices, close_prices, volume, dates = (
            [],
            [],
            [],
            [],
            [],
            [],
        )

        for i in range(0, len(timestamps) - 1):
            start_timestamp = timestamps[i]
            end_timestamp = timestamps[i + 1]
            full_params_str = (
                f"{params_str}&startTime={start_timestamp}&endTime={end_timestamp}"
            )
            url = f"https://api.binance.com/api/v3/klines?{full_params_str}&limit=1000"
            response = requests.get(url)
            if response.status_code == 200:
                json_data = response.json()
                if len(json_data) > 0:
                    for item in json_data:
                        date_time = naive_utcfromtimestamp(item[0] / 1000)
                        dates.append(date_time)
                        open_prices.append(float(item[1]))
                        high_prices.append(float(item[2]))
                        low_prices.append(float(item[3]))
                        close_prices.append(float(item[4]))
                        volume.append(float(item[5]))
            else:
                raise BinanceDataFetchError(
                    f"Error fetching data between {start_timestamp} and {end_timestamp}. \nResponse: {response.status_code} {response.text} \nMake sure you are using valid pair symbol e.g. `ETHUSDC`, not just ETH"
                )

        df = pd.DataFrame(
            {
                "open": open_prices,
                "high": high_prices,
                "low": low_prices,
                "close": close_prices,
                "volume": volume,
            },
            index=dates,
        )

        # df = clean_time_series_data(df)

        # Each timestamp in `timestamps` besides the first and last entry will be duplicated, so remove
        remove_duplicates_df = df[df.index.duplicated(keep="first") == False]

        return remove_duplicates_df

    def fetch_lending_rates(
        self,
        asset_symbol: str,
        time_bucket: TimeBucket,
        start_at: datetime.datetime,
        end_at: datetime.datetime,
        force_redownload=False,
    ) -> pd.Series:
        """Get daily lending interest rates for a given asset from Binance, resampled to the given time bucket.

        :param asset_symbol:

            List of all valid asset symbols as at 2023-11-06 18:00 UTC:

            asset_names = [
                "BTC", "ETH", "XRP", "BNB", "TRX", "USDT", "LINK", "EOS", "ADA", "ONT",
                "USDC", "ETC", "LTC", "XLM", "XMR", "NEO", "ATOM", "DASH", "ZEC", "MATIC",
                "BUSD", "BAT", "IOST", "VET", "QTUM", "IOTA", "XTZ", "BCH", "RVN", "ZIL",
                "ONE", "ANKR", "CELR", "TFUEL", "IOTX", "HBAR", "FTM", "SXP", "BNT", "DOT",
                "REN", "ALGO", "ZRX", "THETA", "COMP", "KNC", "OMG", "KAVA", "BAND", "DOGE",
                "RLC", "WAVES", "MKR", "SNX", "YFI", "CRV", "SUSHI", "UNI", "MANA", "STORJ",
                "UMA", "JST", "AVAX", "NEAR", "FIL", "TRB", "RSR", "TOMO", "OCEAN", "AAVE",
                "SAND", "CHZ", "ARPA", "COTI", "FET", "TROY", "CHR", "ORN", "NMR", "GRT",
                "STPT", "LRC", "KSM", "ROSE", "REEF", "STMX", "ALPHA", "STX", "ENJ", "RUNE",
                "SKL", "INJ", "OXT", "CTSI", "OGN", "EGLD", "1INCH", "DODO", "LIT", "NKN",
                "MDT", "CKB", "CAKE", "SOL", "XEM", "LINA", "GLM", "XVS", "MDX", "SUPER",
                "GTC", "PUNDIX", "AUDIO", "BOND", "SLP", "TRU", "POND", "ERN", "ATA", "NULS",
                "DENT", "TVK", "DF", "FLOW", "AR", "DYDX", "MASK", "UNFI", "AXS", "LUNA",
                "SHIB", "ENS", "BAKE", "ALICE", "TLM", "ICP", "C98", "GALA", "ONG", "HIVE",
                "DAR", "IDEX", "ANT", "CLV", "WAXP", "BNX", "KLAY", "MINA", "XEC", "RNDR",
                "JASMY", "QUICK", "LPT", "AGLD", "BICO", "CTXC", "DUSK", "HOT", "SFP", "YGG",
                "FLUX", "ICX", "CELO", "BETA", "BLZ", "MTL", "PEOPLE", "QNT", "PYR", "KEY",
                "PAXG", "FRONT", "TWT", "RAD", "QI", "GMT", "APE", "BSW", "KDA", "MBL", "ASTR",
                "API3", "CTK", "WOO", "GAL", "OP", "REI", "LEVER", "LDO", "FIDA", "FLM", "BURGER",
                "AUCTION", "IMX", "SPELL", "STG", "BEL", "WING", "AVA", "LOKA", "LUNC", "PHB",
                "LOOM", "AMB", "SANTOS", "VIB", "EPX", "HARD", "USTC", "DEGO", "HIGH", "GMX",
                "LAZIO", "PORTO", "ACH", "STRAX", "KP3R", "REQ", "POLYX", "APT", "PHA", "OSMO",
                "GLMR", "MAGIC", "HOOK", "AGIX", "HFT", "CFX", "ZEN", "SSV", "LQTY", "ALCX",
                "FXS", "PERP", "TUSD", "USDP", "GNS", "JOE", "RIF", "SYN", "ID", "ARB", "OAX",
                "RDNT", "EDU", "SUI", "FLOKI", "PEPE", "COMBO", "MAV", "XVG", "PENDLE", "ARKM",
                "WLD", "T", "FDUSD", "RPL", "SEI", "CYBER", "VTHO", "WBETH", "NTRN", "HIFI",
                "CVX", "ARK", "ARDR", "ACA", "VIDT", "GHST", "GAS", "OOKI", "TIA", "POWR",
                "AERGO", "SNT", "STEEM", "MEME", "PLA", "MULTI", "UFT", "ILV"
            ]

            To see current list of all valid asset symbols, submit API request https://api1.binance.com/sapi/v1/margin/allAssets with your Binance API key.

        :param start_date:
            Start date for the data. Note this value cannot be eariler than datetime.datetime(2019,4,1) due to Binance data limitations

        """
        if not force_redownload:
            try:
                return self.get_data_parquet(
                    asset_symbol, time_bucket, start_at, end_at, is_lending=True
                )
            except:
                pass

        series = self._fetch_lending_rates(asset_symbol, start_at, end_at, time_bucket)

        path = self.get_parquet_path(
            asset_symbol, time_bucket, start_at, end_at, is_lending=True
        )
        series.to_frame(name="lending_rates").to_parquet(path)

        return series

    def _fetch_lending_rates(
        self,
        asset_symbol: str,
        start_at: datetime.datetime,
        end_at: datetime.datetime,
        time_bucket: TimeBucket,
    ) -> pd.Series:
        assert type(asset_symbol) == str, "asset_symbol must be a string"
        assert (
            type(start_at) == datetime.datetime
        ), "start_date must be a datetime.datetime object"
        assert (
            type(end_at) == datetime.datetime
        ), "end_date must be a datetime.datetime object"
        assert (
            type(time_bucket) == TimeBucket
        ), "time_delta must be a pandas Timedelta object"
        # assert start_date >= datetime.datetime(2019,4,1), "start_date cannot be earlier than 2019-04-01 due to Binance data limitations"

        monthly_timestamps = generate_monthly_timestamps(start_at, end_at)
        response_data = []

        # API calls to get the data
        for i in range(len(monthly_timestamps) - 1):
            start_timestamp = monthly_timestamps[i] * 1000
            end_timestamp = monthly_timestamps[i + 1] * 1000
            url = f"https://www.binance.com/bapi/margin/v1/public/margin/vip/spec/history-interest-rate?asset={asset_symbol}&vipLevel=0&size=90&startTime={start_timestamp}&endTime={end_timestamp}"
            response = requests.get(url)
            if response.status_code == 200:
                json_data = response.json()
                data = json_data["data"]
                if len(data) > 0:
                    response_data.extend(data)
            else:
                raise BinanceDataFetchError(
                    f"No data found for {asset_symbol} between {start_at} and {end_at}. Check your symbol matches with valid symbols in method description. \nResponse: {response.status_code} {response.text}"
                )

        dates = []
        interest_rates = []
        for data in response_data:
            dates.append(pd.to_datetime(data["timestamp"], unit="ms"))
            interest_rates.append(float(data["dailyInterestRate"]))

        unsampled_rates = pd.Series(data=interest_rates, index=dates).sort_index()

        # doesn't always raise error
        if unsampled_rates.empty:
            raise BinanceDataFetchError(
                f"No data found for {asset_symbol} between {start_at} and {end_at}. Check your symbol matches with valid symbols in method description. \nResponse: {response.status_code} {response.text}"
            )

        resampled_rates = resample_series(
            unsampled_rates, time_bucket.to_pandas_timedelta(), forward_fill=True
        )

        return resampled_rates

    def fetch_approx_asset_trading_start_date(self, symbol) -> datetime.datetime:
        """Get the asset trading start date at Binance.

        Binance was launched around 2017-08-01.

        :raise BinanceDataFetchError:
            If the asset does not exist.
        """

        monthly_candles = self.fetch_candlestick_data(
            symbol,
            TimeBucket.d30,
            datetime.datetime(2017, 1, 1),
            naive_utcnow(),
        )

        assert len(monthly_candles) > 0, f"Could not find starting date for asset {symbol}"
        return monthly_candles.index[0].to_pydatetime()

    def get_data_parquet(
        self,
        symbol: str,
        time_bucket: TimeBucket,
        start_at: datetime.datetime,
        end_at: datetime.datetime,
        is_lending: bool = False,
    ) -> pd.DataFrame:
        """Get parquet file for the candlestick data.

        :param symbol: Trading pair symbol E.g. ETHUSDC
        :param time_bucket: TimeBucket instance
        :param start_at: Start date of the data
        :param end_at: End date of the data
        :return: Path to the parquet file
        """
        path = self.get_parquet_path(symbol, time_bucket, start_at, end_at, is_lending)
        try:
            return pd.read_parquet(path)
        except Exception as e:
            raise e

    def get_parquet_path(
        self,
        symbol: str,
        time_bucket: TimeBucket,
        start_at: datetime.datetime,
        end_at: datetime.datetime,
        is_lending: bool = False,
    ) -> Path:
        """Get parquet path for the candlestick data.

        :param symbol: Trading pair symbol E.g. ETHUSDC
        :param time_bucket: TimeBucket instance
        :param start_at: Start date of the data
        :param end_at: End date of the data
        :return: Path to the parquet file
        """
        if is_lending:
            file_str = "lending"
        else:
            file_str = "candles"

        file = Path(
            file_str + f"-{symbol}-{time_bucket.value}-{start_at}-{end_at}.parquet"
        )
        return self.cache_directory.joinpath(file)

    def purge_cached_file(
        self,
        *,
        symbol: str = None,
        time_bucket: TimeBucket = None,
        start_at: datetime.datetime = None,
        end_at: datetime.datetime = None,
        path: Path = None,
    ) -> None:
        """Purge specific cached candle data file.

        :param symbol: Trading pair symbol E.g. ETHUSDC
        :param time_bucket: TimeBucket instance
        :param start_at: Start date of the data
        :param end_at: End date of the data
        :param path: Path to the parquet file. If not specified, it will be generated from the other parameters.
        """
        if not path:
            path = self.get_parquet_path(symbol, time_bucket, start_at, end_at)
        if path.exists():
            path.unlink()
        else:
            logger.warn(f"File {path} does not exist.")

    def purge_all_cached_data(self) -> None:
        """Purge all cached candle data. This delete all contents of a cache directory, but not the directory itself. I.e. the cache directory will be left empty

        :param path: Path to the parquet file
        """
        for item in self.cache_directory.iterdir():
            if item.is_dir():
                # Recursively delete directories
                shutil.rmtree(item)
            else:
                # Delete files
                item.unlink()


def convert_binance_lending_rates_to_supply(
    interestRates: pd.Series, multiplier: float = 0.95
) -> pd.Series:
    """Convert Binance lending rates to supply rates.

    Right now, this rate is somewhat arbitrary. It is 95% of the lending rate by default.

    :param interestRates: Series of lending interest rates
    :return: Series of supply rates
    """

    assert 0 < multiplier < 1, "Multiplier must be between 0 and 1"

    assert isinstance(
        interestRates, pd.Series
    ), f"Expected pandas Series, got {interestRates.__class__}: {interestRates}"
    assert isinstance(
        interestRates.index, pd.DatetimeIndex
    ), f"Expected DateTimeIndex, got {interestRates.index.__class__}: {interestRates.index}"
    return interestRates * multiplier


def clean_time_series_data(df: pd.DataFrame | pd.Series) -> pd.DataFrame | pd.Series:
    """Unused for now since data from Binance data occasionally has gaps. Not a huge deal.

    Cleans time series data to ensure:
    - No Nan values
    - Index contains no duplicates
    - Has equally spaced intervals with no gaps
    - Sorted index in ascending order by datetime

    :param df: Pandas dataframe or series
    :return: Cleaned dataframe or series
    """

    if df.isna().any(axis=None):
        raise ValueError("Dataframe contains NaN values")

    if df.duplicated().any():
        raise ValueError("Dataframe contains duplicate values")

    assert type(df.index) == pd.DatetimeIndex, "Index must be a DatetimeIndex"

    df.sort_index(inplace=True)

    if len(uneven_indices := get_indices_of_uneven_intervals(df)) > 0:
        raise ValueError(
            f"Dataframe contains uneven intervals at indices {uneven_indices}"
        )

    return df


def get_indices_of_uneven_intervals(df: pd.DataFrame | pd.Series) -> bool:
    """Checks if a time series contains perfectly evenly spaced time intervals with no gaps.

    :param df: Pandas dataframe or series
    :return: True if time series is perfectly evenly spaced, False otherwise
    """
    assert type(df.index) == pd.DatetimeIndex, "Index must be a DatetimeIndex"

    numeric_representation = df.index.astype(np.int64)

    differences = np.diff(numeric_representation)

    not_equal_to_first = differences != differences[0]

    return np.where(not_equal_to_first)[0]


def get_binance_interval(bucket: TimeBucket) -> str:
    """Convert our TimeBucket to Binance's internal format."""
    if bucket == TimeBucket.d30:
        # Can be one of `1s, 1m, 3m, 5m, 15m, 30m, 1h, 2h, 4h, 6h, 8h, 12h, 1d, 3d, 1w, 1M`
        return "1M"
    else:
        return bucket.value


"""Helpers to timestamp format and value conformity.

- We are operating on naive Python datetimes, all in UTC timezone

"""
import calendar
import datetime

import pandas as pd

#: Pre-instiated no difference pd.Tiemdelta for optimisation
ZERO_TIMEDELTA = pd.Timedelta(0)

def is_compatible_timestamp(ts: pd.Timestamp) -> bool:
    """Ensure Pandas Timestamp is naive.

    We do not carry timezone information in data, because it would slow
    us down.
    """
    assert isinstance(ts, pd.Timestamp), f"We assume pandas.Timestamp, but received {ts}"
    return ts.tz is None


def assert_compatible_timestamp(ts: pd.Timestamp):
    """Check we do not get in bad input timestamps.

    :raise: AssertionError if the timestamp object is incompatible
    """
    assert isinstance(ts, pd.Timestamp), f"not pd.Timestamp: {ts.__class__.__name__}={ts}"
    assert is_compatible_timestamp(ts), f"Received pandas.Timestamp in incompatible format: {type(ts)}: {ts}"


def to_int_unix_timestamp(dt: datetime.datetime) -> int:
    """Get datetime as UTC seconds since epoch."""
    # https://stackoverflow.com/a/5499906/315168
    return int(calendar.timegm(dt.utctimetuple()))


def generate_monthly_timestamps(start: datetime.datetime, end: datetime.datetime) -> list[int]:
    """Generate timestamps from the start to the end. One timestamp per month.
    
    :param start: Start date
    :param end: End date

    :return: List of timestamps
    """
    # TODO: ensure index has no missing dates i.e. evenly spaced intervals throughout the period
    timestamps = []
    current_date = start
    while current_date <= end:
        timestamps.append(int(current_date.timestamp()))
        # Check if adding one month stays within the year
        if current_date.month == 12:
            current_date = current_date.replace(year=current_date.year + 1, month=1)
        else:
            current_date = current_date.replace(month=current_date.month + 1)

        if current_date > end:
            timestamps.append(int(end.timestamp()))
            break

    return timestamps


def naive_utcnow() -> datetime.datetime:
    """Get utcnow() but without timezone information.

    Fixes for Python 3.12 compatibility

    - https://blog.miguelgrinberg.com/post/it-s-time-for-a-change-datetime-utcnow-is-now-deprecated
    """
    return datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)


def naive_utcfromtimestamp(timestamp: float | int) -> datetime.datetime:
    """Get naive UTC datetime from UNIX time.

    Fixes for Python 3.12 compatibility

    - https://blog.miguelgrinberg.com/post/it-s-time-for-a-change-datetime-utcnow-is-now-deprecated
    """
    return datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc).replace(tzinfo=None)
